from __future__ import annotations

from collections import defaultdict
from functools import reduce
import re
from typing import Any, Dict, List, Tuple, cast

from pandas import DataFrame
import pandas as pd

from relationalai import debugging
from relationalai.clients.result_helpers import format_columns, format_value, merge_columns, convert_hash_value, \
    sort_data_frame_result
from relationalai.tools.constants import Generation

# Convert LQP results into the expected single wide table dataframe for the end user
# - Requires identifying and unrolling all GNF relations and populating any resulting nulls
# - Relies on expected ordering of results, as we do not have IDs or names associated with columns here.
# - At the end, col names are rewritten with the expected output names blindly; we trust the stitching
#   has been done correctly for the names to line up with the correct columns
# `result_cols`: the expected output column names in order
# `key_locations`: where output keys are located in outputs
def format_results(results, result_cols:List[str], key_locations:List[int])  -> Tuple[DataFrame, List[Any]]:
    with debugging.span("format_results"):
        data_frame = DataFrame()
        problems = defaultdict(
            lambda: {
                "message": "",
                "path": "",
                "start_line": None,
                "end_line": None,
                "report": "",
                "code": "",
                "severity": "",
                "decl_id": "",
                "name": "",
                "output": "",
                "end_character": None,
                "start_character": None,
                "props": {},
            }
        )

        # Check if there are any results to process
        if not results.results:
            return (data_frame, list(problems.values()))

        out_keys_n = len(key_locations) # number of keys in output
        assert out_keys_n <= len(result_cols)
        out_vals_n = len(result_cols) - out_keys_n # number of values in output

        # only create cols for values, we handle keys separately as they are not GNF
        has_cols:List[DataFrame] = [DataFrame() for _ in range(0, out_vals_n)]
        keys_data_frame = DataFrame()
        key_len = 0

        for result in results.results:
            relation_id = result["relationId"]
            assert isinstance(relation_id, str)
            result_frame = result["table"].to_pandas()
            types = [
                t
                for t in relation_id.split("/")
                if t != "" and not t.startswith(":")
            ]

            # Process diagnostics
            if "/:rel/:catalog/:diagnostic/" in relation_id:
                # Handle different types of diagnostics based on relation_id
                if "/:message/" in relation_id:
                    for _, row in result_frame.iterrows():
                        if len(row) > 1:
                            problems[row.iloc[0]]["message"] = row.iloc[1]
                elif "/:report/" in relation_id:
                    for _, row in result_frame.iterrows():
                        if len(row) > 1:
                            problems[row.iloc[0]]["report"] = row.iloc[1]
                elif "/:start/:line" in relation_id:
                    for _, row in result_frame.iterrows():
                        if len(row) > 2:
                            problems[row.iloc[0]]["start_line"] = row.iloc[2]
                elif "/:end/:line" in relation_id:
                    for _, row in result_frame.iterrows():
                        if len(row) > 2:
                            problems[row.iloc[0]]["end_line"] = row.iloc[2]
                elif "/:model" in relation_id:
                    for _, row in result_frame.iterrows():
                        if len(row) > 1:
                            problems[row.iloc[0]]["path"] = row.iloc[1]
                elif "/:severity" in relation_id:
                    for _, row in result_frame.iterrows():
                        if len(row) > 1:
                            problems[row.iloc[0]]["severity"] = row.iloc[1]
                elif "/:code" in relation_id:
                    for _, row in result_frame.iterrows():
                        if len(row) > 1:
                            problems[row.iloc[0]]["code"] = row.iloc[1]

            # Process integrity constraint violations
            elif "/:rel/:catalog/:ic_violation" in relation_id:
                if "/:decl_id" in relation_id:
                    for _, row in result_frame.iterrows():
                        if len(row) > 1:
                            problems[convert_hash_value(row.iloc[0])]["decl_id"] = (
                                row.iloc[1]
                            )
                            problems[convert_hash_value(row.iloc[0])]["code"] = (
                                "IC_VIOLATION"
                            )
                elif "/:model" in relation_id:
                    for _, row in result_frame.iterrows():
                        if len(row) > 1:
                            problems[convert_hash_value(row.iloc[0])]["path"] = (
                                row.iloc[1]
                            )
                elif "/:name" in relation_id:
                    # Get the last segment of the relation_id as the name
                    segments = [
                        segment[1:]
                        for segment in relation_id.split("/")[4:]
                        if segment.startswith(":")
                    ]
                    for _, row in result_frame.iterrows():
                        problems[convert_hash_value(row.iloc[0])]["name"] = segments[-1]
                elif "/:output" in relation_id:
                    for _, row in result_frame.iterrows():
                        if len(row) > 1:
                            problems[convert_hash_value(row.iloc[0])]["message"] = (
                                row.iloc[1]
                            )
                elif "/:range/:end/:character" in relation_id:
                    for _, row in result_frame.iterrows():
                        if len(row) > 1:
                            problems[convert_hash_value(row.iloc[0])][
                                "end_character"
                            ] = row.iloc[1]
                elif "/:range/:end/:line" in relation_id:
                    for _, row in result_frame.iterrows():
                        if len(row) > 1:
                            problems[convert_hash_value(row.iloc[0])]["end_line"] = (
                                row.iloc[1]
                            )
                elif "/:range/:start/:character" in relation_id:
                    for _, row in result_frame.iterrows():
                        if len(row) > 1:
                            problems[convert_hash_value(row.iloc[0])][
                                "start_character"
                            ] = row.iloc[1]
                elif "/:range/:start/:line" in relation_id:
                    for _, row in result_frame.iterrows():
                        if len(row) > 1:
                            problems[convert_hash_value(row.iloc[0])]["start_line"] = (
                                row.iloc[1]
                            )
                elif "/:report" in relation_id:
                    for _, row in result_frame.iterrows():
                        if len(row) > 1:
                            problems[convert_hash_value(row.iloc[0])]["report"] = (
                                row.iloc[1]
                            )

            elif "/:pyrel_error_attrs" in relation_id:
                assert len(result_frame.columns) == 3, \
                    f"Error attribute with {len(result_frame.columns)} cols, expected 3"
                for _, row in result_frame.iterrows():
                    id = convert_hash_value(row.iloc[0])
                    problems[id]["code"] = "PYREL_ERROR"
                    if row.iloc[1] == "message":
                        problems[id]["message"] = row.iloc[2]
                    elif row.iloc[1] == "severity":
                        problems[id]["severity"] = row.iloc[2]
                    else:
                        props = cast(Dict, problems[id]["props"])
                        props[row.iloc[1]] = format_value(row.iloc[2], relation_id.split("/")[-1])

            elif "/:__pyrel_debug_watch" in relation_id:
                result_frame = format_columns(result_frame, types, Generation.QB)
                from relationalai.experimental.inspect import _print_watch_frame
                _print_watch_frame(result_frame)

            # Process outputs
            else:
                result_frame = format_columns(result_frame, types, Generation.QB)
                result["table"] = result_frame
                matched = re.search(r"^/:output_cols_col([0-9]+)/", relation_id)
                if matched:
                    col_ix = int(matched.group(1))

                    # Generate col names and write them into the df (idn for keys, vn for cols)
                    key_cols = [f"id{i}" for i in range(0, len(result_frame.columns) - 1)]
                    key_len = len(key_cols)
                    # Don't use the actual name as users are free to use our generated names
                    result_col_name = f"v{col_ix}"
                    result_frame.columns = [*key_cols, result_col_name]

                    has_cols[col_ix] = result_frame
                elif relation_id.startswith("/:output_keys"):
                    # data for all keys (wide), to merge in later
                    keys_data_frame = result_frame

                    # Rename wide key col names to match key cols in df_wide_reset
                    keys_data_frame.columns = pd.RangeIndex(len(keys_data_frame.columns))
                    keys_data_frame = keys_data_frame.rename(columns=lambda c: f"id{c}")

                elif relation_id.startswith("/:output"): # wide outputs case
                    data_frame = pd.concat(
                        [data_frame, result_frame], ignore_index=True
                    )

        # GNF values case: stitch together output vals and keys into one wide dataframe
        if any(not col.empty for col in has_cols):
            # Merge value cols together by their key cols
            key_cols = [f"id{i}" for i in range(0, key_len)]
            df_wide_reset = reduce(lambda left, right: merge_columns(left, right, key_cols), has_cols)

            # Join wide keys with wide vals (keys all at the front; still needs reordering)
            data_frame = pd.merge(keys_data_frame, df_wide_reset, on=key_cols, how='outer')

            # Reorder outputs
            if key_locations:
                data_frame = _shift_keys(data_frame, keys_data_frame, out_keys_n, key_locations)

            else: # if no keys in output, just drop all of the key cols
                data_frame = data_frame.drop(columns=key_cols)

        # Empty values case: reorder/drop keys as needed
        elif not keys_data_frame.empty:
            if key_locations: # Reorder outputs
                data_frame = _shift_keys(keys_data_frame, keys_data_frame, out_keys_n, key_locations)
            else: # if there are no keys to output, we may still need to populate nulls for output values
                # Take into account the cols that could contain values (even though they're empty)
                key_cols = [f"id{i}" for i in range(0, len(keys_data_frame.columns))]
                has_cols.append(keys_data_frame) # include the keys so we know how many nulls to generate
                df_wide_reset = reduce(lambda left, right: merge_columns(left, right, key_cols), has_cols)
                data_frame = df_wide_reset.drop(columns=key_cols)

        data_frame = sort_data_frame_result(data_frame)

        # Overwrite column names with user-defined names
        # The assumption is that the extra keys have been chopped off the front, and the
        # remaining columns are in the correct order and require renaming
        if 0 < len(data_frame.columns) <= len(result_cols):
            data_frame.columns = result_cols[:len(data_frame.columns)]

        return (data_frame, list(problems.values()))

# Reorder `res` df to match user-specified output order and drop non-output key cols
# E.g., Current df looks like:
#   [out_key_pos1, out_key_pos4, hidden_key_1, value_pos2, value_pos3, value_pos5]
# Target df looks like:
#   [out_key_pos1, value_pos2, value_pos3, out_key_pos4, value_pos5]
def _shift_keys(res:DataFrame, keys_data_frame:DataFrame, out_keys_n:int, key_locations:List[int]):
    offset = len(keys_data_frame.columns) # index of first value in data frame
    assert out_keys_n <= offset
    extra_keys_n = offset - out_keys_n # number of keys to drop (those not in output)

    # Shift output keys into the correct spots and drop the rest
    for i in key_locations:
        res = res[res.columns[1:].insert(i + offset - 1, res.columns[0])]
        offset -= 1
    res = res.drop(columns=res.columns[:extra_keys_n])
    return res



from .discharge_constraints import DischargeConstraints
from .dnf_union_splitter import DNFUnionSplitter
from .format_outputs import FormatOutputs
from .handle_aggregations_and_ranks import HandleAggregationsAndRanks

__all__ = ["DischargeConstraints", "DNFUnionSplitter", "FormatOutputs", "HandleAggregationsAndRanks"]



from collections import defaultdict
from typing import Optional, TypeGuard, Union, cast
from relationalai.semantics.metamodel import ir, helpers, factory, builtins
from relationalai.semantics.metamodel.compiler import Pass
from relationalai.semantics.metamodel.visitor import Visitor, Rewriter, collect_by_type
from relationalai.semantics.lqp.algorithms import (
    is_script, is_algorithm_script,is_logical_instruction, is_update_instruction,
    get_instruction_head_rels, get_instruction_body_rels, mk_assign, mk_debug, split_instruction,
    get_update_instruction_arity
)
from relationalai.semantics.lqp.builtins import has_debug_annotation, get_debug_annotation

class AlgorithmPass(Pass):
    """
    Transforms algorithm scripts by normalizing Loopy constructs (iterative algorithm).

    This pass applies three main rewriting transformations to Metamodel IR that prepare
    algorithm scripts for execution, in the order listed below:

    1. *Intermediate Rescoping*: Moves nested logical intermediate relations from their
       original logical scope into algorithm scripts, placing them immediately before each
       instruction that uses them (which can include Break instructions). Removes
       intermediates from the logical scope if they're only used within algorithms.
       TODO: Monitor https://github.com/RelationalAI/relationalai-python/pull/3187

       Example (Metamodel IR):
           BEFORE:
               Logical
                   Logical
                       R(x::Int128, y::Int128)
                       → derive _nested_logical_1(x::Int128, y::Int128) @assign
                   Sequence @script @algorithm
                       Logical
                           _nested_logical_1(a::Int128, b::Int128)
                           → derive S(a::Int128, b::Int128) @assign

           AFTER:
               Logical
                   Sequence @script @algorithm
                       Logical
                           R(x::Int128, y::Int128)
                           → derive _nested_logical_1(x::Int128, y::Int128) @assign
                       Logical
                           _nested_logical_1(a::Int128, b::Int128)
                           → derive S(a::Int128, b::Int128) @assign

    2. **Update Normalization**: Transforms Loopy update operations (@upsert, @monoid, @monus)
       to use a single body atom. Complex bodies with multiple lookups or additional
       operations are normalized by introducing intermediate relations.

       Example (Metamodel IR):
           BEFORE:
               Logical
                   R(x::Int128, y::Int128)
                   S(y::Int128, z::Int128)
                   → derive T(x::Int128, z::Int128) @upsert

           AFTER:
               Logical
                   R(x::Int128, y::Int128)
                   S(y::Int128, z::Int128)
                   → derive _loopy_update_intermediate_1(x::Int128, z::Int128) @assign

               Logical
                   _loopy_update_intermediate_1(x::Int128, z::Int128)
                   → derive T(x::Int128, z::Int128) @upsert

    3. **Recursive Assignment Decoupling**: Decouples self-referential assignments where the
       head relation appears in the body by introducing a copy relation. This transformation
       is required for BackIR analysis compatibility.

       Example (Metamodel IR):
           BEFORE:
               Logical
                   iter(i::Int128)
                   rel_primitive_int128_add(i::Int128, 1::Int128, i_plus_1::Int128)
                   → derive iter(i_plus_1::Int128) @assign

           AFTER:
               Logical
                   iter(i::Int128)
                   → derive _loopy_iter_copy_1(i::Int128) @assign

               Logical
                   _loopy_iter_copy_1(i::Int128)
                   rel_primitive_int128_add(i::Int128, 1::Int128, i_plus_1::Int128)
                   → derive iter(i_plus_1::Int128) @assign

    4. **Debug Annotation Propagation** Uses dependency analysis to propagate any debugging annotations
        that originate from @break rules down to their dependencies.

        Example:
            BEFORE:
                Logical
                    → derive foo(2) @assign
                Break @recursion_logging("recursion_content", 1)
                    foo(i::Int128)
                    i = 2

            AFTER:
                Logical
                    → derive foo(2) @assign @recursion_logging("recursion_content", 1)
                Break @recursion_logging("recursion_content", 1)
                    foo(i::Int128)
                    i = 2
    """
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        # Find all nested logical intermediates
        intermediate_finder = FindIntermediates()
        model.accept(intermediate_finder)

        intermediate_analyzer = AnalyzeIntermediateUse(set(intermediate_finder.intermediates.keys()))
        model.accept(intermediate_analyzer)

        # Determine which intermediates to move and which to remove
        uses_intermediates: dict[Union[ir.Logical, ir.Break], set[ir.Logical]] = defaultdict(set)
        remove_declarations: set[ir.Logical] = set()
        for rel, decl in intermediate_finder.intermediates.items():
            if rel not in intermediate_analyzer.used_outside_algorithm:
                remove_declarations.add(decl)
            for instr in intermediate_analyzer.used_in_alg_instruction[rel]:
                uses_intermediates[instr].add(decl)

        # Rescope intermediates
        rescoper = IntermediateRescoper(uses_intermediates, remove_declarations)
        model = rescoper.walk(model)

        # Normalize Loopy updates
        normalizer = UpdateNormalizer()
        model = normalizer.walk(model)

        # Decompose recursive assignments
        decomposer = RecursiveAssignmentDecoupling()
        model = decomposer.walk(model)

        # Propagate debug annotations
        propagator = PropagateDebugAnnos()
        model = propagator.walk(model)

        return model

class FindIntermediates(Visitor):
    """
    Gathers all `_nested_logical.*` intermediates defined in a Logical scope (where order
    doesn't matter); in particular DOES NOT gather any intermediates declared in the scope
    of a Sequence.
    """
    def __init__(self):
        self.intermediates: dict[ir.Relation, ir.Logical] = dict()
        self._inside_algorithm: bool = False

    def visit_logical(self, node: ir.Logical, parent: Optional[ir.Node]):
        if is_logical_instruction(node):
            heads = get_instruction_head_rels(node)
            for rel in heads:
                if rel.name.startswith("_nested_logical"):
                    self.intermediates[rel] = node
        else:
            super().visit_logical(node, parent)

    def visit_sequence(self, node: ir.Sequence, parent: Optional[ir.Node]):
        if is_algorithm_script(node):
            self._inside_algorithm = True
        super().visit_sequence(node, parent)
        if is_algorithm_script(node):
            self._inside_algorithm = False


class AnalyzeIntermediateUse(Visitor):
    """
    Identifies, for each nested logical intermediate, the algorithm instructions that
    use it. Additionally, determines whether the intermediate is used anywhere
    outside of an algorithm.
    """
    def __init__(self, intermediate_relations: set[ir.Relation]):
        self.intermediates = intermediate_relations
        self.used_in_algorithm: set[ir.Relation] = set()
        self.used_in_alg_instruction: dict[ir.Relation, set[Union[ir.Logical, ir.Break]]] = {rel: set() for rel in intermediate_relations}
        self.used_outside_algorithm: set[ir.Relation] = set()

        self._current_algorithm: Optional[ir.Sequence] = None

    def register_use(self, instr: Union[ir.Logical, ir.Break], uses_intermediates: set[ir.Relation]):
        # this instruction uses intermediates
        if self._current_algorithm is not None:
            # instruction is inside an algorithm
            for rel in uses_intermediates:
                self.used_in_algorithm.add(rel)
                self.used_in_alg_instruction[rel].add(instr)
        else:
            self.used_outside_algorithm.update(uses_intermediates)

    def visit_break(self, node: ir.Break, parent: Optional[ir.Node]):
        lookups = collect_by_type(ir.Lookup, node)
        lookup_rels = {lookup.relation for lookup in lookups}
        uses_intermediates = lookup_rels.intersection(self.intermediates)
        self.register_use(node, uses_intermediates)
        super().visit_break(node, parent)

    def visit_logical(self, node: ir.Logical, parent: Optional[ir.Node]):
        if is_logical_instruction(node):
            body = get_instruction_body_rels(node)
            uses_intermediates = body.intersection(self.intermediates)
            self.register_use(node, uses_intermediates)
        else:
            super().visit_logical(node, parent)

    def visit_sequence(self, node: ir.Sequence, parent: Optional[ir.Node]):
        if is_algorithm_script(node):
            self._current_algorithm = node
        super().visit_sequence(node, parent)
        if is_algorithm_script(node):
            self._current_algorithm = None

class IntermediateRescoper(Rewriter):
    """
    Moves nested logical intermediates used in algorithm instructions from the logical scope
    to any algorithm using the instruction before every instruction that uses them. Removes
    an intermediate from the logical scope if it is not used anywhere else.

    * `uses_intermediates`: a mapping from algorithm instructions to the set of nested logical
      intermediates they use.
    * `remove_declarations`: the set of nested logical intermediates to remove from the
      logical scope because they are not used anywhere else.
    """
    def __init__(self,
                 uses_intermediates: dict[Union[ir.Logical, ir.Break], set[ir.Logical]],
                 remove_declarations: set[ir.Logical]):
        super().__init__()
        self.uses_intermediates = uses_intermediates
        self.remove_declarations = remove_declarations

    def handle_logical(self, node: ir.Logical, parent: ir.Node) -> ir.Logical:
        body = []
        for child in node.body:
            if child in self.remove_declarations:
                continue
            child = self.walk(child, node)
            body.append(child)
        return node.reconstruct(node.engine, node.hoisted, tuple(body), node.annotations)

    def handle_sequence(self, node: ir.Sequence, parent: ir.Node) -> ir.Sequence:
        tasks = []
        for child in node.tasks:
            if child in self.uses_intermediates:
                assert isinstance(child, (ir.Logical, ir.Break))
                for intermediate in self.uses_intermediates[child]:
                    tasks.append(mk_assign(intermediate))
            child = self.walk(child, node)
            tasks.append(child)
        return node.reconstruct(node.engine, node.hoisted, tuple(tasks), node.annotations)

class UpdateNormalizer(Rewriter):
    """
    This pass normalizes Loopy Update operations (upsert, monoid, and monus) to use a single
    atom in their body. For any Update operation with more complex body, it introduces a new
    intermediate relation to hold the body results.
    """
    def __init__(self):
        super().__init__()
        self._inside_algorithm: bool = False
        self._intermediate_counter: int = 0

    # Tests if the given Update operation requires normalization
    # * the body has more than one Lookup operation, or
    # * the body has other tasks than Lookup and Update
    def _requires_update_normalization(self, update: ir.Task) -> bool:
        if not isinstance(update, ir.Logical):
            return False
        if not is_update_instruction(update):
            return False
        _, lookups, others = split_instruction(update)
        # Special case: even a single lookup needs normalization if its relation
        # is handled specially by _translate_to_atom (producing non-Atom LQP nodes).
        if len(lookups) == 1:
            rel = lookups[0].relation
            if rel in (builtins.cast, builtins.join, builtins.infomap,
                       builtins.louvain, builtins.label_propagation):
                return True
            if builtins.is_pragma(rel) or builtins.is_builtin(rel) or helpers.is_external(rel):
                return True
        return len(lookups) > 1 or len(others) > 0

    def handle_sequence(self, node: ir.Sequence, parent: ir.Node) -> ir.Sequence:
        if is_algorithm_script(node):
            self._inside_algorithm = True

        if self._inside_algorithm:
            new_tasks = []
            for task in node.tasks:
                if self._requires_update_normalization(task):
                    assert isinstance(task, ir.Logical)
                    intermediate, normalized_update = self._normalize_update_instruction(task)
                    new_tasks.extend((intermediate, normalized_update))
                else:
                    new_tasks.append(self.walk(task, node))
            result = node.reconstruct(node.engine, node.hoisted, tuple(new_tasks), node.annotations)
        else:
            result = super().handle_sequence(node, parent)

        if is_algorithm_script(node):
            self._inside_algorithm = False

        return result

    def _normalize_update_instruction(self, update_instr: ir.Logical) -> tuple[ir.Logical, ir.Logical]:
        update, lookups, others = split_instruction(update_instr)
        normalized_update = []

        var_list = helpers.vars(update.args)

        intermediate_rel = factory.relation(
            self._fresh_intermediate_name(), [
                factory.field(f"arg_{i}", var.type) for i, var in enumerate(var_list)
            ]
        )

        arity = get_update_instruction_arity(update)
        annos = [] if arity == 0 else [builtins.function_checked_annotation]

        intermediate_derive = factory.derive(intermediate_rel, var_list, annos=annos)

        intermediate_logical = mk_assign(factory.logical(
            engine=update_instr.engine,
            hoisted=update_instr.hoisted,
            body=(*lookups, *others, intermediate_derive),
            annos=list(update_instr.annotations)
        ))
        assert isinstance(intermediate_logical, ir.Logical)

        intermediate_lookup = factory.lookup(
            intermediate_rel,
            var_list
        )

        normalized_update = factory.logical(
            engine=update_instr.engine,
            hoisted=update_instr.hoisted,
            body=(intermediate_lookup, update),
            annos=list(update_instr.annotations)
        )

        return (intermediate_logical, normalized_update)

    def _fresh_intermediate_name(self) -> str:
        self._intermediate_counter += 1
        return f"_loopy_update_intermediate_{self._intermediate_counter}"

class RecursiveAssignmentDecoupling(Rewriter):
    """
    Decouples assignments whose definition is "recursive", i.e., the body contain the head
    e.g., `assign iter = iter + 1`. Currently, BackIR analysis cannot handle properly such
    assignments. Such assignments are decoupled by introducing a new intermediate copy
    relation; in the example above, `assign iter_copy = iter; assign iter = iter_copy + 1`.
    The performance is not affected because the backend can identify the new assignment as a
    copy operation and the execution will not lead to materialization of the intermediate
    relation.
    """
    def __init__(self):
        super().__init__()
        self._intermediate_copy_counter: int = 0
        # control of head_rel -> copy_rel substitution in traversal
        self._perform_substitution: bool = False
        self._head_rel: Optional[ir.Relation] = None
        self._copy_rel: Optional[ir.Relation] = None

    def _fresh_copy_rel_name(self, rel_name:str) -> str:
        self._intermediate_copy_counter += 1
        return f"_loopy_{rel_name}_copy_{self._intermediate_copy_counter}"

    def handle_sequence(self, node: ir.Sequence, parent: ir.Node) -> ir.Sequence:
        if is_script(node):
            new_tasks = []
            for task in node.tasks:
                if self._is_recursive_assignment(task):
                    assert isinstance(task, ir.Logical)
                    intermediate_copy, decomposed_assign = self._decouple_recursive_assignment(task, parent)
                    new_tasks.extend((intermediate_copy, decomposed_assign))
                else:
                    new_tasks.append(self.walk(task, node))
            return node.reconstruct(node.engine, node.hoisted, tuple(new_tasks), node.annotations)
        else:
            return super().handle_sequence(node, parent)

    def _is_recursive_assignment(self, task: ir.Task) -> TypeGuard[ir.Logical]:
        if is_logical_instruction(task):
            heads = get_instruction_head_rels(task)
            body = get_instruction_body_rels(task)
            return len(body & heads) > 0
        return False

    def _decouple_recursive_assignment(self, rule: ir.Logical, parent: ir.Node) -> tuple[ir.Logical, ir.Logical]:
        # we have `assign rel(x,...) = ..., rel(y,...), ...`
        update, _, _ = split_instruction(rule)
        self._head_rel = update.relation

        copy_rel_name = self._fresh_copy_rel_name(self._head_rel.name)

        self._copy_rel = factory.relation(copy_rel_name, list(self._head_rel.fields))
        # build `assign copy_rel(x,...) = rel(x,...)`
        copy_rule = cast(ir.Logical, mk_assign(
            factory.logical([
                factory.lookup(self._head_rel,update.args),
                factory.update(self._copy_rel, update.args, update.effect)
            ])
        ))

        # build `assign rel(x,...) = ..., copy_rel(y,...), ...``
        self._perform_substitution = True
        rewritten_rule = self.walk(rule, parent)
        self._perform_substitution = False

        self._head_rel = None
        self._copy_rel = None

        return (copy_rule, rewritten_rule)

    def handle_lookup(self, node: ir.Lookup, parent: ir.Node) -> ir.Lookup:
        if self._perform_substitution and node.relation == self._head_rel:
            assert self._copy_rel is not None
            return factory.lookup(self._copy_rel, node.args)
        return super().handle_lookup(node, parent)

class PropagateDebugAnnos(Rewriter):
    """
    Propagates debug annotations from Break nodes to the Update instructions they depend on.

    For each Break with a @recursion_logging annotation, we find the relations it reads
    (via Lookups in its check) and then annotate sibling Logical instructions in the same
    script that write to those relations.
    """
    def __init__(self):
        super().__init__()

    def handle_sequence(self, node: ir.Sequence, parent: ir.Node) -> ir.Sequence:
        if not is_script(node):
            return super().handle_sequence(node, parent)

        # Collect debug annotations from Break nodes, keyed by the relations they read
        debug_annos_by_rel: dict[ir.Relation, list[ir.Annotation]] = defaultdict(list)
        for task in node.tasks:
            if isinstance(task, ir.Break) and has_debug_annotation(task):
                anno = get_debug_annotation(task)
                if anno is None:
                    continue
                for lookup in collect_by_type(ir.Lookup, task):
                    debug_annos_by_rel[lookup.relation].append(anno)

        if not debug_annos_by_rel:
            return super().handle_sequence(node, parent)

        # Rewrite tasks: propagate annotations to instructions that write to those relations
        new_tasks = []
        for task in node.tasks:
            task = self.walk(task, node)
            if is_logical_instruction(task):
                head_rels = get_instruction_head_rels(task)
                for rel in head_rels:
                    if rel in debug_annos_by_rel:
                        for anno in debug_annos_by_rel[rel]:
                            task = mk_debug(task, anno)
            new_tasks.append(task)

        return node.reconstruct(node.engine, node.hoisted, tuple(new_tasks), node.annotations)



"""Pretty-printing for LQP IR Transactions."""
from __future__ import annotations

import re

from lqp.gen.pretty import pretty, pretty_debug

from relationalai.semantics.lqp.lqp_bundle.ir import Transaction
from relationalai.semantics.lqp.lqp_bundle.proto_conversion import _transaction_to_proto


def to_safe_string(
    txn: Transaction,
    *,
    debug: bool = False,
) -> str:
    """Pretty-print an IR Transaction to a string, with CSV filenames hidden."""
    try:
        proto = _transaction_to_proto(txn)
        result = pretty_debug(proto) if debug else pretty(proto)
        result = re.sub(
            r'"snowflake://[^"]+/export/export_[^"/]+/data"',
            '"<hidden filename>"',
            result,
        )
        return result
    except Exception as exc:
        return f"<lqp_print error: {repr(exc)}>"



from relationalai.semantics.metamodel.compiler import Pass
from relationalai.semantics.metamodel import ir, builtins as rel_builtins, factory as f, visitor
from relationalai.semantics.metamodel.typer import typer
from relationalai.semantics.metamodel import helpers, types
from relationalai.semantics.metamodel.util import FrozenOrderedSet, OrderedSet
from relationalai.semantics.lqp.utils import extract_export_fqn


from typing import cast, Union, Optional, Iterable
from collections import defaultdict

# LQP does not support multiple definitions for the same relation. This pass unifies all
# definitions for each relation into a single definition using a union.
class UnifyDefinitions(Pass):
    def __init__(self):
        super().__init__()

    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        # Maintain a cache of renamings for each relation. These need to be consistent
        # across all definitions of the same relation.
        self.renamed_relation_args: dict[Union[ir.Value, ir.Relation], list[ir.Var]] = {}

        root = cast(ir.Logical, model.root)
        new_tasks = self.get_combined_multidefs(root)
        return ir.Model(
            model.engines,
            model.relations,
            model.types,
            f.logical(
                tuple(new_tasks),
                root.hoisted,
                root.engine,
            ),
            model.annotations,
        )

    def _get_heads(self, logical: ir.Logical) -> list[Union[ir.Update, ir.Output]]:
        derives = []
        for task in logical.body:
            if isinstance(task, ir.Update) and task.effect == ir.Effect.derive:
                derives.append(task)
            elif isinstance(task, ir.Output):
                derives.append(task)
        return derives

    def _get_non_heads(self, logical: ir.Logical) -> list[ir.Task]:
        non_derives = []
        for task in logical.body:
            if not(isinstance(task, ir.Update) and task.effect == ir.Effect.derive) and not isinstance(task, ir.Output):
                non_derives.append(task)
        return non_derives

    def _get_head_identifier(self, head: Union[ir.Update, ir.Output]) -> Optional[ir.Value]:
        if isinstance(head, ir.Update):
            return head.relation
        else:
            assert isinstance(head, ir.Output)
            if len(head.aliases) < 2:
                # For processing here, we need output to have at least the column marker
                # `keys` or both `cols` and `col` markers, and also a key
                return None

            output_alias_names = helpers.output_alias_names(head.aliases)
            output_vals = helpers.output_values(head.aliases)

            # For normal outputs, the pattern is output[keys](cols, "col000" as 'col', ...)
            if output_alias_names[0] == "cols" and output_alias_names[1] == "col":
                return output_vals[1]
            if output_alias_names[0] == "keys": # handle wide keys relation
                return output_vals[0]

            # For exports, the pattern is output[keys]("col000" as 'col', ...)
            # We want to unify exports based on both the export FQN and the column marker since 
            # different exports could have the same column marker but represent different relations.
            if helpers.is_export(head):
                if output_alias_names[0] == "col":
                    # Use a composite identifier: "export_fqn::column_marker"
                    # This distinguishes both different exports AND different columns
                    export_fqn_str = extract_export_fqn(head.annotations)
                    column_marker = output_vals[0]

                    # Create string-based composite identifier
                    if export_fqn_str:
                        # Convert column_marker to string representation
                        if isinstance(column_marker, ir.Literal):
                            col_str = str(column_marker.value)
                        else:
                            col_str = str(column_marker)
                        return ir.Literal(types.String, f"{export_fqn_str}::{col_str}")
                    # Fallback to just column marker if no FQN found
                    return column_marker

        return None

    def get_combined_multidefs(self, root: ir.Logical) -> list[ir.Logical]:
        # Step 1: Group tasks by the relation they define.
        relation_to_tasks: dict[Union[None, ir.Value, ir.Relation], list[ir.Logical]] = defaultdict(list)

        for task in root.body:
            task = cast(ir.Logical, task)
            task_heads = self._get_heads(task)

            # Some relations do not need to be grouped, e.g., if they don't contain a
            # derive. Use `None` as a placeholder key for these cases.
            if len(task_heads) != 1:
                relation_to_tasks[None].append(task)
                continue

            head_id = self._get_head_identifier(task_heads[0])
            relation_to_tasks[head_id].append(task)

        # Step 2: For each relation, combine all of the body definitions into a union.
        result_tasks = []
        for relation, tasks in relation_to_tasks.items():
            # If there's only one task for the relation, or if grouping is not needed, then
            # just keep the original tasks.
            if len(tasks) == 1 or relation is None:
                result_tasks.extend(tasks)
                continue

            result_tasks.append(self._combine_tasks_into_union(tasks))
        return result_tasks

    def _get_variable_mapping(self, logical: ir.Logical) -> dict[ir.Value, ir.Var]:
        heads = self._get_heads(logical)
        assert len(heads) == 1, "should only have one head in a logical at this stage"
        head = heads[0]

        var_mapping = {}
        head_id = self._get_head_identifier(head)

        if isinstance(head, ir.Update):
            args_for_renaming = head.args
        else:
            assert isinstance(head, ir.Output)
            output_alias_names = helpers.output_alias_names(head.aliases)
            if output_alias_names[0] == "cols" and output_alias_names[1] == "col":
                assert len(head.aliases) > 2

                # For outputs, we do not need to rename the `cols` and `col` markers or the
                # keys.
                output_values = helpers.output_values(head.aliases)[2:]

            elif output_alias_names[0] == "keys": # handle wide keys output
                assert len(head.aliases) > 1
                output_values = helpers.output_values(head.aliases)[1:]

            else:
                assert helpers.is_export(head) and output_alias_names[0] == "col"
                assert len(head.aliases) > 1

                # For exports, we do not need to rename the `col` marker or the keys.
                output_values = helpers.output_values(head.aliases)[1:]

            args_for_renaming = []
            for v in output_values:
                if head.keys and isinstance(v, ir.Var) and v in head.keys:
                    continue
                args_for_renaming.append(v)

        if head_id not in self.renamed_relation_args:
            renamed_vars = []
            for (i, arg) in enumerate(args_for_renaming):
                typ = typer.to_type(arg)
                assert arg not in var_mapping, "args of update should be unique"
                if isinstance(arg, ir.Var):
                    var_mapping[arg] = ir.Var(typ, arg.name)
                else:
                    var_mapping[arg] = ir.Var(typ, f"arg_{i}")

                renamed_vars.append(var_mapping[arg])
            self.renamed_relation_args[head_id] = renamed_vars
        else:
            for (arg, var) in zip(args_for_renaming, self.renamed_relation_args[head_id]):
                var_mapping[arg] = var

        return var_mapping

    def _rename_variables(self, logical: ir.Logical) -> ir.Logical:
        class RenameVisitor(visitor.Rewriter):
            def __init__(self, var_mapping: dict[ir.Value, ir.Var]):
                super().__init__()
                self.var_mapping = var_mapping

            def _get_mapped_value(self, val: ir.Value) -> ir.Value:
                if isinstance(val, tuple):
                    return tuple(self._get_mapped_value(t) for t in val)
                return self.var_mapping.get(val, val)

            def _get_mapped_values(self, vals: Iterable[ir.Value]) -> list[ir.Value]:
                return [self._get_mapped_value(v) for v in vals]

            def handle_var(self, node: ir.Var, parent: ir.Node) -> ir.Var:
                return self.var_mapping.get(node, node)

            # TODO: ideally, extend the rewriter class to allow rewriting PyValue to Var so
            # we don't need to separately handle all cases containing them.
            def handle_update(self, node: ir.Update, parent: ir.Node) -> ir.Update:
                return ir.Update(
                    node.engine,
                    node.relation,
                    tuple(self._get_mapped_values(node.args)),
                    node.effect,
                    node.annotations,
                )

            def handle_lookup(self, node: ir.Lookup, parent: ir.Node) -> ir.Lookup:
                return ir.Lookup(
                    node.engine,
                    node.relation,
                    tuple(self._get_mapped_values(node.args)),
                    node.annotations,
                )

            def handle_output(self, node: ir.Output, parent: ir.Node) -> ir.Output:
                new_aliases = FrozenOrderedSet(
                    [(name, self._get_mapped_value(value)) for name, value in node.aliases]
                )
                if node.keys:
                    new_keys = FrozenOrderedSet(
                        [self.var_mapping.get(key, key) for key in node.keys]
                    )
                else:
                    new_keys = node.keys

                return ir.Output(
                    node.engine,
                    new_aliases,
                    new_keys,
                    node.annotations,
                )

            def handle_construct(self, node: ir.Construct, parent: ir.Node) -> ir.Construct:
                new_values = tuple(self._get_mapped_values(node.values))
                new_id_var = self.var_mapping.get(node.id_var, node.id_var)
                return ir.Construct(
                    node.engine,
                    new_values,
                    new_id_var,
                    node.annotations,
                )

            def handle_aggregate(self, node: ir.Aggregate, parent: ir.Node) -> ir.Aggregate:
                new_projection = tuple(self.var_mapping.get(arg, arg) for arg in node.projection)
                new_group = tuple(self.var_mapping.get(arg, arg) for arg in node.group)
                new_args = tuple(self._get_mapped_values(node.args))
                return ir.Aggregate(
                    node.engine,
                    node.aggregation,
                    new_projection,
                    new_group,
                    new_args,
                    node.annotations,
                )

            def handle_rank(self, node: ir.Rank, parent: ir.Node) -> ir.Rank:
                new_projection = tuple(self.var_mapping.get(arg, arg) for arg in node.projection)
                new_group = tuple(self.var_mapping.get(arg, arg) for arg in node.group)
                new_args = tuple(self.var_mapping.get(arg, arg) for arg in node.args)
                new_result = self.var_mapping.get(node.result, node.result)

                return ir.Rank(
                    node.engine,
                    new_projection,
                    new_group,
                    new_args,
                    node.arg_is_ascending,
                    new_result,
                    node.limit,
                    node.annotations,
                )

        var_mapping = self._get_variable_mapping(logical)

        renamer = RenameVisitor(var_mapping)
        result = renamer.walk(logical)

        # Also need to append the equality for each renamed constant. E.g., if the mapping
        # contains (50.0::FLOAT -> arg_2::FLOAT), we need to add
        # `eq(arg_2::FLOAT, 50.0::FLOAT)` to the result.
        value_eqs = []
        for (old_var, new_var) in var_mapping.items():
            if not isinstance(old_var, ir.Var):
                value_eqs.append(f.lookup(rel_builtins.eq, [new_var, old_var]))

        return ir.Logical(
            result.engine,
            result.hoisted,
            tuple(value_eqs) + tuple(result.body),
            result.annotations,
        )

    # This function is the main workhorse for this rewrite pass. It takes a list of tasks
    # that define the same relation, and combines them into a single task that defines
    # the relation using a union of all of the bodies.
    def _combine_tasks_into_union(self, tasks: list[ir.Logical]) -> ir.Logical:
        # Step 1: Rename the variables in all tasks so that they will match the final derive
        # after reconstructing into a union
        renamed_tasks = [self._rename_variables(task) for task in tasks]

        # Step 2: Get the final derive
        derives = self._get_heads(renamed_tasks[0])
        assert len(derives) == 1, "should only have one derive in a logical at this stage"
        # Also make sure that all the derives are the same. This should be the case because
        # we renamed all the variables to be the same in step 1.
        for task in renamed_tasks[1:]:
            assert self._get_heads(task) == derives, "all derives should be the same"

        derive = derives[0]

        # Step 3: Remove the final `derive` from each task
        renamed_task_bodies = [
            f.logical(
                tuple(self._get_non_heads(t)),  # Only keep non-head tasks
                t.hoisted,
                t.engine,
            )
            for t in renamed_tasks
        ]

        # Deduplicate bodies
        renamed_task_bodies = OrderedSet.from_iterable(renamed_task_bodies).get_list()

        # Step 4: Construct a union of all the task bodies
        if len(renamed_task_bodies) == 1:
            # If there's only one body after deduplication, no need to create a union
            new_body = renamed_task_bodies[0]
        else:
            new_body = f.union(
                tuple(renamed_task_bodies),
                [],
                renamed_tasks[0].engine,
            )

        # Step 5: Add the final derive back
        return f.logical(
            (new_body, derive),
            [],
            renamed_tasks[0].engine,
        )
